package ru.dbotthepony.kstarbound.lua

import com.sun.jna.Callback
import com.sun.jna.Library
import com.sun.jna.Native
import com.sun.jna.Pointer
import com.sun.jna.ptr.LongByReference
import org.apache.logging.log4j.LogManager
import org.lwjgl.system.MemoryUtil
import java.io.Closeable
import java.io.PrintWriter
import java.io.StringWriter
import java.lang.RuntimeException
import java.nio.ByteBuffer
import java.nio.ByteOrder

private typealias size_t = Long
private typealias psize_t = LongByReference

private typealias lua_KContext = Pointer

interface LuaJNADynamic : Library {
	fun lua_atpanic(luaState: Pointer, fn: lua_CFunction): Pointer
	fun lua_load(luaState: Pointer, reader: lua_Reader, data: Pointer?, chunkName: String, mode: String): Int
	fun lua_pushlstring(luaState: Pointer, str: Pointer, len: size_t): Pointer?

	interface lua_CFunction : Callback {
		fun invoke(luaState: Pointer): Int
	}

	interface lua_Reader : Callback {
		fun invoke(luaState: Pointer, data: Pointer?, size: psize_t): Pointer? /* String */
	}

	interface lua_KFunction : Callback {
		/*
		** Type for continuation functions
		*/
		fun invoke(luaState: Pointer, status: Int, ctx: lua_KContext?): Int
	}
}

val LUA_JNA: LuaJNADynamic = Native.load("lua54", LuaJNADynamic::class.java)

private fun stringToBuffer(str: String): ByteBuffer {
	val bytes = str.toByteArray(charset = Charsets.UTF_8)
	val buf = ByteBuffer.allocateDirect(bytes.size)
	buf.order(ByteOrder.nativeOrder())
	bytes.forEach(buf::put)
	buf.position(0)

	return buf
}

private val SHARED_BUFFER = ByteBuffer.allocateDirect(16384).also { it.order(ByteOrder.nativeOrder()) }
private val SHARED_BUFFER_PTR = Pointer(MemoryUtil.memAddress(SHARED_BUFFER))

private fun loadString(string: String): Int {
	SHARED_BUFFER.position(0)
	val bytes = string.toByteArray(charset = Charsets.UTF_8)
	bytes.forEach(SHARED_BUFFER::put)
	SHARED_BUFFER.position(0)
	return bytes.size
}

private fun lua_pushlstring(luaState: Pointer, str: String) {
	val len = loadString(str)
	LUA_JNA.lua_pushlstring(luaState, SHARED_BUFFER_PTR, len.toLong())
}

class LuaState : Closeable {
	private val pointer = LuaJNA.luaL_newstate()
	private val nativePointer = Pointer.nativeValue(pointer)

	private var destroyed = false
	private val panicHandler = object : LuaJNADynamic.lua_CFunction {
		override fun invoke(luaState: Pointer): Int {
			throw RuntimeException("$this panicked!")
		}
	}

	init {
		if (pointer == Pointer.NULL)
			throw OutOfMemoryError("Unable to allocate new LuaState")

		LUA_JNA.lua_atpanic(pointer, panicHandler)

		LuaJNA.luaopen_base(nativePointer)
		LuaJNA.luaopen_package(nativePointer)
		LuaJNA.luaopen_table(nativePointer)
		LuaJNA.luaopen_coroutine(nativePointer)
		LuaJNA.luaopen_string(nativePointer)
		LuaJNA.luaopen_math(nativePointer)
		LuaJNA.luaopen_utf8(nativePointer)
		LuaJNA.luaopen_debug(nativePointer)

		pushClosure {
			val build = mutableListOf<String?>()

			for (i in 1 .. stackTop) {
				build.add(getString(i))
			}

			LOGGER.info("Lua/print(): {}", build.joinToString("\t"))
		}

		LuaJNA.lua_setglobal(nativePointer, "print")
	}

	fun pushClosure(lambda: (state: LuaState) -> Unit) {
		LuaJNI.lua_pushcclosure(nativePointer) lazy@{
			try {
				lambda.invoke(this@LuaState)
			} catch (err: Throwable) {
				val builder = StringWriter()
				val printWriter = PrintWriter(builder)
				err.printStackTrace(printWriter)
				lua_pushlstring(pointer, builder.toString())
				return@lazy 1
			}

			return@lazy 0
		}
	}

	fun checkStack(minAmount: Int): Boolean {
		return LuaJNA.lua_checkstack(nativePointer, minAmount) > 0
	}

	fun absoluteIndex(index: Int): Int {
		return LuaJNA.lua_absindex(nativePointer, index)
	}

	fun load(code: String, chunkName: String = "main chunk") {
		val buf = stringToBuffer(code)

		throwLoadError(LUA_JNA.lua_load(pointer, object : LuaJNADynamic.lua_Reader {
			override fun invoke(luaState: Pointer, data: Pointer?, size: psize_t): Pointer? {
				if (buf.remaining() == 0) {
					size.value = 0
					return Pointer.NULL
				}

				size.value = buf.remaining().toLong()
				val p = Pointer(MemoryUtil.memAddress(buf))
				buf.position(buf.remaining())
				return p
			}
		}, Pointer.NULL, chunkName, "t"))
	}

	fun pcall(numArgs: Int = 0, numResults: Int = 0): Int {
		val status = LuaJNA.lua_pcallk(nativePointer, numArgs, numResults, 0, null, null)

		if (status == LUA_ERRRUN) {
			throw LuaRuntimeException(getString())
		}

		return status
	}

	fun getString(index: Int = -1, limit: Long = 4096): String? {
		val len = psize_t()
		val p = LuaJNA.lua_tolstring(nativePointer, absoluteIndex(index), len) ?: return null

		if (len.value == 0L) {
			return ""
		}

		if (len.value >= limit) {
			throw IllegalStateException("Unreasonably long Lua string: ${len.value}")
		}

		val readBytes = ByteArray(len.value.toInt())
		p.read(0L, readBytes, 0, readBytes.size)
		return readBytes.toString(charset = Charsets.UTF_8)
	}

	val stackTop get() = LuaJNA.lua_gettop(nativePointer)

	override fun close() {
		if (destroyed) {
			throw IllegalStateException("Already destroyed")
		}

		LuaJNA.lua_close(nativePointer)
		destroyed = true
	}

	companion object {
		private val LOGGER = LogManager.getLogger(LuaState::class.java)
	}
}
